import struct
from dataclasses import dataclass

import factom_core.primitives as primitives
from factom_core.utils import varint


class AdminMessage:
    """Base class to inherit from"""

    ADMIN_ID = None

    def marshal(self):
        raise NotImplementedError("Must be implemented by subclasses")

    @classmethod
    def unmarshal(cls, raw: bytes):
        raise NotImplementedError("Must be implemented by subclasses")

    def to_dict(self):
        raise NotImplementedError("Must be implemented by subclasses")


def unmarshal_admin_message_of_type(admin_id: int, data: bytes) -> AdminMessage:
    """Given an admin ID type and the marshalled message bytes, return the unmarshalled message."""
    if admin_id == MinuteNumber.ADMIN_ID:
        return MinuteNumber.unmarshal(data)
    elif admin_id == DirectoryBlockSignature.ADMIN_ID:
        return DirectoryBlockSignature.unmarshal(data)
    elif admin_id == MatryoshkaHashReveal.ADMIN_ID:
        return MatryoshkaHashReveal.unmarshal(data)
    elif admin_id == MatryoshkaHashAddOrReplace.ADMIN_ID:
        return MatryoshkaHashAddOrReplace.unmarshal(data)
    elif admin_id == ServerCountIncrease.ADMIN_ID:
        return ServerCountIncrease.unmarshal(data)
    elif admin_id == AddFederatedServer.ADMIN_ID:
        return AddFederatedServer.unmarshal(data)
    elif admin_id == AddAuditServer.ADMIN_ID:
        return AddAuditServer.unmarshal(data)
    elif admin_id == RemoveFederatedServer.ADMIN_ID:
        return RemoveFederatedServer.unmarshal(data)
    elif admin_id == AddFederatedServerSigningKey.ADMIN_ID:
        return AddFederatedServerSigningKey.unmarshal(data)
    elif admin_id == AddFederatedServerBitcoinAnchorKey.ADMIN_ID:
        return AddFederatedServerBitcoinAnchorKey.unmarshal(data)
    elif admin_id == ServerFaultHandoff.ADMIN_ID:
        return ServerFaultHandoff()
    elif admin_id == CoinbaseDescriptor.ADMIN_ID:
        return CoinbaseDescriptor.unmarshal(data)
    elif admin_id == CoinbaseDescriptorCancel.ADMIN_ID:
        return CoinbaseDescriptorCancel.unmarshal(data)
    elif admin_id == AddAuthorityFactoidAddress.ADMIN_ID:
        return AddAuthorityFactoidAddress.unmarshal(data)
    elif admin_id == AddAuthorityEfficiency.ADMIN_ID:
        return AddAuthorityEfficiency.unmarshal(data)
    else:
        raise ValueError("Invalid Admin ID")


@dataclass
class MinuteNumber(AdminMessage):
    """Minute marker (Deprecated in M2)"""

    ADMIN_ID = 0x00
    MESSAGE_SIZE = 1

    minute: int

    def __post_init__(self):
        # TODO: value assertions
        assert 1 <= self.minute <= 10, "minute must be in range(1, 11)"

    def marshal(self):
        """
        Marshal the object into its byte representation:
        - single byte encoding the minute number

        :return: bytes representation of the MinuteNumber message
        """
        buf = bytearray()
        buf.append(self.minute)
        return bytes(buf)

    @classmethod
    def unmarshal(cls, raw: bytes):
        """
        Unmarshal the byte representation into a new object

        :param raw: marshalled bytes of the message
        :return: new MinuteNumber message object
        """
        minute, data = raw[0], raw[1:]
        assert len(data) == 0, "Extra bytes remaining!"
        return MinuteNumber(minute)

    def to_dict(self):
        return {"type": self.ADMIN_ID, "minute": self.minute}


@dataclass
class DirectoryBlockSignature(AdminMessage):
    """A signature of the preceding Directory Block header."""

    ADMIN_ID = 0x01
    MESSAGE_SIZE = 128

    chain_id: bytes  # the servers 32 byte identity ChainID
    signature: primitives.FullSignature

    def __post_init__(self):
        assert len(self.chain_id) == 32, "chain_id must be a bytes object of length 32"
        assert self.signature is not None, "signature must not be None"

    def marshal(self):
        """
        Marshal the object into its byte representation:
        - the first 32 bytes are the server's identity ChainID
        - next 32 bytes are the Ed25519 public key in that identity
        - next 64 bytes are the signature of the previous Directory Block's header

        :return: bytes representation of the DirectoryBlockSignature message
        """
        buf = bytearray()
        buf.extend(self.chain_id)
        buf.extend(self.signature.marshal())
        return bytes(buf)

    @classmethod
    def unmarshal(cls, raw: bytes):
        """
        Unmarshal the byte representation into a new object

        :param raw: marshalled bytes of the message
        :return: new DirectoryBlockSignature message object
        """
        chain_id, data = raw[:32], raw[32:]
        signature, data = primitives.FullSignature.unmarshal(data[:96]), data[96:]
        assert len(data) == 0, "Extra bytes remaining!"
        return DirectoryBlockSignature(chain_id, signature)

    def to_dict(self):
        return {
            "type": self.ADMIN_ID,
            "chain_id": self.chain_id.hex(),
            "signature": self.signature.to_dict(),
        }


@dataclass
class MatryoshkaHashReveal(AdminMessage):
    """This is the latest M-hash reveal to be considered for determining server priority in subsequent blocks."""

    ADMIN_ID = 0x02
    MESSAGE_SIZE = 64

    chain_id: bytes  # the servers 32 byte identity ChainID
    matryoshka_hash_reveal: bytes  # 32 bytes for the M-hash reveal itself

    def __post_init__(self):
        assert len(self.chain_id) == 32, "chain_id must be a bytes object of length 32"
        assert len(self.matryoshka_hash_reveal) == 32, "matryoshka_hash_reveal must be a bytes object of length 32"

    def marshal(self):
        """
        Marshal the object into its byte representation:
        - the first 32 bytes are the server's identity ChainID
        - next 32 bytes for the new M-hash reveal itself

        :return: bytes representation of the MatryoshkaHashReveal message
        """
        buf = bytearray()
        buf.extend(self.chain_id)
        buf.extend(self.matryoshka_hash_reveal)
        return bytes(buf)

    @classmethod
    def unmarshal(cls, raw: bytes):
        """
        Unmarshal the byte representation into a new object

        :param raw: marshalled bytes of the message
        :return: new MatryoshkaHashReveal message object
        """
        chain_id, data = raw[:32], raw[32:]
        matryoshka_hash_reveal, data = data[:32], data[32:]
        assert len(data) == 0, "Extra bytes remaining!"
        return MatryoshkaHashReveal(chain_id, matryoshka_hash_reveal)

    def to_dict(self):
        return {
            "type": self.ADMIN_ID,
            "chain_id": self.chain_id.hex(),
            "matryoshka_hash_reveal": self.matryoshka_hash_reveal.hex(),
        }


@dataclass
class MatryoshkaHashAddOrReplace(AdminMessage):
    """
    This is a command which adds or replaces the current M-hash for the specified identity with this new M-hash.
    This data is replicated from the server's identity chain.

    """

    ADMIN_ID = 0x03
    MESSAGE_SIZE = 64

    chain_id: bytes  # the servers 32 byte identity ChainID
    new_matryoshka_hash: bytes  # 32 bytes for the new M-hash itself

    def __post_init__(self):
        assert len(self.chain_id) == 32, "chain_id must be a bytes object of length 32"
        assert len(self.new_matryoshka_hash) == 32, "new_matryoshka_hash must be a bytes object of length 32"

    def marshal(self):
        """
        Marshal the object into its byte representation:
        - the first 32 bytes are the server's identity ChainID
        - next 32 bytes for the new M-hash itself

        :return: bytes representation of the MatryoshkaHashAddOrReplace message
        """
        buf = bytearray()
        buf.extend(self.chain_id)
        buf.extend(self.new_matryoshka_hash)
        return bytes(buf)

    @classmethod
    def unmarshal(cls, raw: bytes):
        """
        Unmarshal the byte representation into a new object

        :param raw: marshalled bytes of the message
        :return: new MatryoshkaHashAddOrReplace message object
        """
        chain_id, data = raw[:32], raw[32:]
        new_matryoshka_hash, data = data[:32], data[32:]
        assert len(data) == 0, "Extra bytes remaining!"
        return MatryoshkaHashAddOrReplace(chain_id, new_matryoshka_hash)

    def to_dict(self):
        return {
            "type": self.ADMIN_ID,
            "chain_id": self.chain_id.hex(),
            "new_matryoshka_hash": self.new_matryoshka_hash.hex(),
        }


@dataclass
class ServerCountIncrease(AdminMessage):
    """This is a command to increase the number of servers."""

    ADMIN_ID = 0x04
    MESSAGE_SIZE = 1

    value: int  # the server count is incremented by this amount

    def __post_init__(self):
        assert 0 <= self.value <= 255, "value must be in range(0, 256)"

    def marshal(self):
        """
        Marshal the object into its byte representation:
        - single byte encoding the amount to increase the count by

        :return: bytes representation of the ServerCountIncrease message
        """
        buf = bytearray()
        buf.append(self.value)
        return bytes(buf)

    @classmethod
    def unmarshal(cls, raw: bytes):
        """
        Unmarshal the byte representation into a new object

        :param raw: marshalled bytes of the message
        :return: new ServerCountIncrease message object
        """
        value, data = raw[0], raw[1:]
        assert len(data) == 0, "Extra bytes remaining!"
        return ServerCountIncrease(value)

    def to_dict(self):
        return {"type": self.ADMIN_ID, "value": self.value}


@dataclass
class AddFederatedServer(AdminMessage):
    """Add the specified identity to the pool of Federated Servers"""

    ADMIN_ID = 0x05
    MESSAGE_SIZE = 36

    chain_id: bytes  # the servers 32 byte identity ChainID
    activation_height: int  # the Directory Block height that it takes effect

    def __post_init__(self):
        assert len(self.chain_id) == 32, "chain_id must be a bytes object of length 32"
        assert self.activation_height >= 0, "activation_height must not be negative"

    def marshal(self):
        """
        Marshal the object into its byte representation:
        - the first 32 bytes are the ChainID of the Federated server which is added to the pool
        - next 4 bytes are the Directory Block height that it takes effect

        :return: bytes representation of the AddFederatedServer message
        """
        buf = bytearray()
        buf.extend(self.chain_id)
        buf.extend(struct.pack(">I", self.activation_height))
        return bytes(buf)

    @classmethod
    def unmarshal(cls, raw: bytes):
        """
        Unmarshal the byte representation into a new object

        :param raw: marshalled bytes of the message
        :return: new AddFederatedServer message object
        """
        chain_id, data = raw[:32], raw[32:]
        activation_height, data = struct.unpack(">I", data[:4])[0], data[4:]
        assert len(data) == 0, "Extra bytes remaining!"
        return AddFederatedServer(chain_id, activation_height)

    def to_dict(self):
        return {
            "type": self.ADMIN_ID,
            "chain_id": self.chain_id.hex(),
            "activation_height": self.activation_height,
        }


@dataclass
class AddAuditServer(AdminMessage):
    """Add the specified identity to the pool of Audit Servers"""

    ADMIN_ID = 0x06
    MESSAGE_SIZE = 36

    chain_id: bytes  # the servers 32 byte identity ChainID
    activation_height: int  # the Directory Block height that it takes effect

    def __post_init__(self):
        assert len(self.chain_id) == 32, "chain_id must be a bytes object of length 32"
        assert self.activation_height >= 0, "activation_height must not be negative"

    def marshal(self):
        """
        Marshal the object into its byte representation:
        - the first 32 bytes are the ChainID of the Audit server which is added to the pool
        - next 4 bytes are the Directory Block height that it takes effect

        :return: bytes representation of the AddAuditServer message
        """
        buf = bytearray()
        buf.extend(self.chain_id)
        buf.extend(struct.pack(">I", self.activation_height))
        return bytes(buf)

    @classmethod
    def unmarshal(cls, raw: bytes):
        """
        Unmarshal the byte representation into a new object

        :param raw: marshalled bytes of the message
        :return: new AddAuditServer message object
        """
        chain_id, data = raw[:32], raw[32:]
        activation_height, data = struct.unpack(">I", data[:4])[0], data[4:]
        assert len(data) == 0, "Extra bytes remaining!"
        return AddAuditServer(chain_id, activation_height)

    def to_dict(self):
        return {
            "type": self.ADMIN_ID,
            "chain_id": self.chain_id.hex(),
            "activation_height": self.activation_height,
        }


@dataclass
class RemoveFederatedServer(AdminMessage):
    """
    Remove the specified identity from the pool of Federated Servers (also can be used to remove an Audit server)
    All public keys associated with it are removed as well.
    """

    ADMIN_ID = 0x07
    MESSAGE_SIZE = 36

    chain_id: bytes  # the servers 32 byte identity ChainID
    activation_height: int  # the Directory Block height that it takes effect

    def __post_init__(self):
        assert len(self.chain_id) == 32, "chain_id must be a bytes object of length 32"
        assert self.activation_height >= 0, "activation_height must not be negative"

    def marshal(self):
        """
        Marshal the object into its byte representation:
        - the first 32 bytes are the ChainID of the Federated server which is removed from the pool
        - next 4 bytes are the Directory Block height that it takes effect

        :return: bytes representation of the RemoveFederatedServer message
        """
        buf = bytearray()
        buf.extend(self.chain_id)
        buf.extend(struct.pack(">I", self.activation_height))
        return bytes(buf)

    @classmethod
    def unmarshal(cls, raw: bytes):
        """
        Unmarshal the byte representation into a new object

        :param raw: marshalled bytes of the message
        :return: new RemoveFederatedServer message object
        """
        chain_id, data = raw[:32], raw[32:]
        activation_height, data = struct.unpack(">I", data[:4])[0], data[4:]
        assert len(data) == 0, "Extra bytes remaining!"
        return RemoveFederatedServer(chain_id, activation_height)

    def to_dict(self):
        return {
            "type": self.ADMIN_ID,
            "chain_id": self.chain_id.hex(),
            "activation_height": self.activation_height,
        }


@dataclass
class AddFederatedServerSigningKey(AdminMessage):
    """
    This adds an Ed25519 public key to the specified identity in the authority set.
    If the specified key for this server already exists, this replaces the old one.
    """

    ADMIN_ID = 0x08
    MESSAGE_SIZE = 69

    chain_id: bytes  # the servers 32 byte identity ChainID
    priority: int  # the key's priority level
    new_public_key: bytes  # the new 32 byte Ed25519 public key
    activation_height: int  # the Directory Block height that it takes effect

    def __post_init__(self):
        assert len(self.chain_id) == 32, "chain_id must be a bytes object of length 32"
        assert 0 <= self.priority <= 255, "priority must be in range(0, 256)"
        assert len(self.new_public_key) == 32, "new_public_key must be a bytes object of length 32"
        assert self.activation_height >= 0, "activation_height must not be negative"

    def marshal(self):
        """
        Marshal the object into its byte representation:
        - the first 32 bytes are the server's identity ChainID
        - next byte is the key priority
        - next 32 bytes are the public key itself
        - next 4 bytes are the activation height

        :return: bytes representation of the AddFederatedServerSigningKey message
        """
        buf = bytearray()
        buf.extend(self.chain_id)
        buf.append(self.priority)
        buf.extend(self.new_public_key)
        buf.extend(struct.pack(">I", self.activation_height))
        return bytes(buf)

    @classmethod
    def unmarshal(cls, raw: bytes):
        """
        Unmarshal the byte representation into a new object

        :param raw: marshalled bytes of the message
        :return: new AddFederatedServerSigningKey message object
        """
        chain_id, data = raw[:32], raw[32:]
        priority, data = data[0], data[1:]
        new_public_key, data = data[:32], data[32:]
        activation_height, data = struct.unpack(">I", data[:4])[0], data[4:]
        assert len(data) == 0, "Extra bytes remaining!"
        return AddFederatedServerSigningKey(chain_id, priority, new_public_key, activation_height)

    def to_dict(self):
        return {
            "type": self.ADMIN_ID,
            "chain_id": self.chain_id.hex(),
            "priority": self.priority,
            "new_public_key": self.new_public_key,
            "activation_height": self.activation_height,
        }


@dataclass
class AddFederatedServerBitcoinAnchorKey(AdminMessage):
    """
    This adds a Bitcoin public key hash to the authority set.
    If the specified priority for the server already exists, this replaces the old one.
    """

    ADMIN_ID = 0x09
    MESSAGE_SIZE = 54  # TODO: IS THIS TRUE MESSAGE SIZE?

    chain_id: bytes  # the servers 32 byte identity ChainID
    priority: int
    hash_type: int
    public_key_hash: bytes

    def __post_init__(self):
        assert len(self.chain_id) == 32, "chain_id must be a bytes object of length 32"
        assert self.hash_type == 0 or self.hash_type == 1, "hash_type must be 0 (p2pkh) or 1 (p2sh)"
        assert 0 <= self.priority <= 255, "priority must be in range(0, 256)"
        assert len(self.public_key_hash) == 20, "public_key_hash must be a bytes object of length 20"

    def marshal(self):
        """
        Marshal the object into its byte representation:
        - the first 32 bytes are the server's identity ChainID
        - next byte is the key priority
        - next byte is 0=P2PKH 1=P2SH
        - next 20 bytes are the HASH160 of the ECDSA public key

        :return: bytes representation of the AddFederatedServerBitcoinAnchorKey message
        """
        buf = bytearray()
        buf.extend(self.chain_id)
        buf.append(self.priority)
        buf.append(self.hash_type)
        buf.extend(self.public_key_hash)
        return bytes(buf)

    @classmethod
    def unmarshal(cls, raw: bytes):
        """
        Unmarshal the byte representation into a new object

        :param raw: marshalled bytes of the message
        :return: new AddFederatedServerBitcoinAnchorKey message object
        """
        chain_id, data = raw[:32], raw[32:]
        priority, data = data[0], data[1:]
        hash_type, data = data[0], data[1:]
        public_key_hash, data = data[:20], data[20:]
        assert len(data) == 0, "Extra bytes remaining!"
        return AddFederatedServerBitcoinAnchorKey(chain_id, priority, hash_type, public_key_hash)

    def to_dict(self):
        return {
            "type": self.ADMIN_ID,
            "chain_id": self.chain_id.hex(),
            "priority": self.priority,
            "hash_type": self.hash_type,
            "public_key_hash": self.public_key_hash,
        }


@dataclass
class ServerFaultHandoff(AdminMessage):
    ADMIN_ID = 0x0A
    MESSAGE_SIZE = 0
    """
    This holds a rollup of all the messages which were sent out by the federated servers which authorize
    the removal of one server and the promotion of another server. This is not currently serialized into
    the blockchain.
    """


@dataclass
class CoinbaseDescriptor(AdminMessage):
    """
    The coinbase descriptor is an entry in the Admin block which specifies a number of factoid output addresses and
    amounts. These outputs are used to deterministically generate a coinbase transaction 1000 blocks (about 7 days)
    later. This delay is to allow 7 days to respond to software bugs, etc. It is included in each block height that
    is divisible by 25. There can only be one per Admin block.
    """

    ADMIN_ID = 0x0B
    MAX_MESSAGE_SIZE = 10232

    outputs: list  # the outputs to include in the coinbase transactions 1000 blocks in the future

    def __post_init__(self):
        for output in self.outputs:
            assert (
                "value" in output and "fct_address" in output
            ), "Invalid output! Must contain a value and a fct_address"
            value, fct_address = output["value"], output["fct_address"]
            assert isinstance(value, int) and value >= 0, "Invalid output! `value` must be a positive integer"
            assert (
                isinstance(fct_address, bytes) and len(fct_address) == 32
            ), "Invalid output! fct_address must be a bytes object of length 32"

    def marshal(self):
        """
        Marshal the object into its byte representation:
        - the first byte is a length descriptor varint (10232 at the maximum)
        - next varint bytes is the descriptor height at which the output will not be created
        - next varint bytes is the index into the specified descriptor that will not be created

        The Coinbase Descriptor cannot be larger than 10233 bytes:
        (1 AdminID byte + 2 varint length bytes + (10 KiB max FCT tx - 10 header bytes of coinbase))

        :return: bytes representation of the CoinbaseDescriptor message
        """
        buf = bytearray()
        bodybuf = bytearray()
        for output in self.outputs:
            bodybuf.extend(varint.encode(output.get("value")))
            bodybuf.extend(output.get("fct_address"))
        buf.extend(varint.encode(len(bodybuf)))
        buf.extend(bodybuf)
        return bytes(buf)

    @classmethod
    def unmarshal(cls, raw: bytes):
        """
        Unmarshal the byte representation into a new object

        :param raw: marshalled bytes of the message
        :return: new CoinbaseDescriptor message object
        """
        msg, data = CoinbaseDescriptor.unmarshal_with_remainder(raw)
        assert len(data) == 0, "Extra bytes remaining!"
        return msg

    @classmethod
    def unmarshal_with_remainder(cls, raw: bytes):
        """
        Unmarshal the byte representation into a new object and return the remaining bytes

        :param raw: marshalled bytes of the message
        :return: a tuple of (new CoinbaseDescriptor message object, remaining bytes)
        """
        message_size, data = varint.decode(raw)
        message_data, data = data[:message_size], data[message_size:]
        outputs = []
        while len(message_data) > 0:
            value, message_data = varint.decode(message_data)
            fct_address, message_data = message_data[:32], message_data[32:]
            outputs.append({"value": value, "fct_address": fct_address})
        assert len(message_data) == 0, "Extra bytes remaining in message data!"
        return CoinbaseDescriptor(outputs), data

    def to_dict(self):
        return {}  # TODO: CoinbaseDescriptor to_dict


@dataclass
class CoinbaseDescriptorCancel(AdminMessage):
    """
    This cancels a specific output index in an earlier Coinbase Descriptor.
    It is only effective in a block between when the Coinbase Descriptor is created and when the coinbase is
    included in the Factoid Block (non-inclusive).
    """

    ADMIN_ID = 0x0C
    MAX_MESSAGE_SIZE = 8

    descriptor_height: int  # the height at which to target the CoinbaseDescriptor
    descriptor_index: int  # this index into the specified descriptor will not be created

    def __post_init__(self):
        assert self.descriptor_height >= 0, "descriptor_height must not be negative"
        assert self.descriptor_index >= 0, "descriptor_index must not be negative"

    def marshal(self):
        """
        Marshal the object into its byte representation:
        - the first byte is a length descriptor varint (8 at the maximum)
        - next varint bytes is the descriptor height at which the output will not be created
        - next varint bytes is the index into the specified descriptor that will not be created

        :return: bytes representation of the CoinbaseDescriptorCancel message
        """
        buf = bytearray()
        bodybuf = bytearray()
        bodybuf.extend(varint.encode(self.descriptor_height))
        bodybuf.extend(varint.encode(self.descriptor_index))
        buf.extend(varint.encode(len(bodybuf)))
        buf.extend(bodybuf)
        return bytes(buf)

    @classmethod
    def unmarshal(cls, raw: bytes):
        """
        Unmarshal the byte representation into a new object

        :param raw: marshalled bytes of the message
        :return: new CoinbaseDescriptorCancel message object
        """
        msg, data = CoinbaseDescriptorCancel.unmarshal_with_remainder(raw)
        assert len(data) == 0, "Extra bytes remaining!"
        return msg

    @classmethod
    def unmarshal_with_remainder(cls, raw: bytes):
        """
        Unmarshal the byte representation into a new object and return the remaining bytes

        :param raw: marshalled bytes of the message
        :return: a tuple of (new CoinbaseDescriptorCancel message object, remaining bytes)
        """
        message_size, data = varint.decode(raw)
        message_data, data = data[:message_size], data[message_size:]
        descriptor_height, message_data = varint.decode(message_data)
        descriptor_index, message_data = message_data[:32], message_data[32:]
        assert len(message_data) == 0, "Extra bytes remaining in message data!"
        return CoinbaseDescriptorCancel(descriptor_height, descriptor_index), data

    def to_dict(self):
        return {}  # TODO: CoinbaseDescriptorCancel to_dict


@dataclass
class AddAuthorityFactoidAddress(AdminMessage):
    """
    This sets a Factoid address to be used in the Coinbase Descriptor.
    If the specified address for this identity already exists, this replaces the old one.
    """

    ADMIN_ID = 0x0D
    MESSAGE_SIZE = 65

    chain_id: bytes  # the server's 32 byte identity ChainID
    fct_address: bytes

    def __post_init__(self):
        assert len(self.chain_id) == 32, "chain_id must be a bytes object of length 32"
        assert len(self.fct_address) == 32, "fct_address must be a bytes object of length 32"

    def marshal(self):
        """
        Marshal the object into its byte representation:
        - the first byte is a length descriptor varint (always 64)
        - next 32 bytes are the server's identity ChainID
        - next 32 bytes are the Factoid address (RCD Hash)

        :return: bytes representation of the AddAuthorityFactoidAddress message
        """
        buf = bytearray()
        buf.append(64)
        buf.extend(self.chain_id)
        buf.extend(self.fct_address)
        return bytes(buf)

    @classmethod
    def unmarshal(cls, raw: bytes):
        """
        Unmarshal the byte representation into a new object

        :param raw: marshalled bytes of the message
        :return: new AddAuthorityFactoidAddress message object
        """
        data = raw[1:]  # Skip message length, always 64
        chain_id, data = data[:32], data[32:]
        fct_address, data = data[:32], data[32:]
        return AddAuthorityFactoidAddress(chain_id, fct_address)

    def to_dict(self):
        return {
            "type": self.ADMIN_ID,
            "chain_id": self.chain_id.hex(),
            "fct_address": self.fct_address,
        }


@dataclass
class AddAuthorityEfficiency(AdminMessage):
    """
    This sets what percentage of the Factoid rewards for the specified server are yeilded to the Grant Pool.
    This command overrides the previous efficiency settings for this identity.
    """

    ADMIN_ID = 0x0E
    MESSAGE_SIZE = 35

    chain_id: bytes  # the server's 32 byte identity ChainID
    efficiency_percentage: int

    def __post_init__(self):
        assert len(self.chain_id) == 32, "chain_id must be a bytes object of length 32"
        assert 0 <= self.efficiency_percentage <= 10000, "efficiency_percentage must be in range(0, 10000)"
        # TODO: is this efficiency assertion right?

    def marshal(self):
        """
        Marshal the object into its byte representation:
        - the first byte is a length descriptor varint (always 34)
        - next 32 bytes are the server's identity ChainID
        - next 2 bytes are a big endian representation of the percentage with 2 fixed decimals

        :return: bytes representation of the AddAuthorityEfficiency message
        """
        buf = bytearray()
        buf.extend(self.chain_id)
        buf.extend(struct.pack(">H", self.efficiency_percentage))
        return bytes(buf)

    @classmethod
    def unmarshal(cls, raw: bytes):
        """
        Unmarshal the byte representation into a new object

        :param raw: marshalled bytes of the message
        :return: new AddAuthorityEfficiency message object
        """
        data = raw[1:]  # Skip message length, always 34
        chain_id, data = data[:32], data[32:]
        efficiency_percentage, data = struct.unpack(">H", data[:4])[0], data[4:]
        return AddAuthorityEfficiency(chain_id, efficiency_percentage)

    def to_dict(self):
        return {
            "type": self.ADMIN_ID,
            "chain_id": self.chain_id.hex(),
            "efficiency": self.efficiency_percentage,
        }



from dataclasses import dataclass
from factom_core.utils import varint


@dataclass
class BalanceIncrease:
    ECID = 0x04

    ec_public_key: bytes
    tx_id: bytes
    index: int
    quantity: int

    def __post_init__(self):
        # TODO: assert they're all here
        pass

    def marshal(self):
        """Marshals the BalanceIncrease according to the byte-level representation shown at
        https://github.com/FactomProject/FactomDocs/blob/master/factomDataStructureDetails.md#balance-increase
        """
        data = self.ec_public_key
        data += self.tx_id
        data += varint.encode(self.index)
        data += varint.encode(self.quantity)
        return data

    @classmethod
    def unmarshal(cls, raw: bytes):
        """Returns a new BalanceIncrease object, unmarshalling given bytes according to:
        https://github.com/FactomProject/FactomDocs/blob/master/factomDataStructureDetails.md#balance-increase
        """
        obj, data = BalanceIncrease.unmarshal_with_remainder(raw)
        assert len(data) == 0, "Extra bytes remaining!"
        return obj

    @classmethod
    def unmarshal_with_remainder(cls, raw: bytes):
        """Returns a new BalanceIncrease object along with any remaining data, unmarshalling given bytes according to:
        https://github.com/FactomProject/FactomDocs/blob/master/factomDataStructureDetails.md#balance-increase

        Because BalanceIncrease is variable length, it is useful to pass a "stream" of bytes to be unmarshalled.
        This way, we don't have to know the size in advance. Just return the remainder bytes for the caller to use
        elsewhere.
        """
        ec_public_key, data = raw[:32], raw[32:]
        tx_id, data = data[:32], data[32:]
        index, data = varint.decode(data)
        quantity, data = varint.decode(data)

        return (
            BalanceIncrease(ec_public_key=ec_public_key, tx_id=tx_id, index=index, quantity=quantity),
            data,
        )

    def to_dict(self):
        return {
            "ec_public_key": self.ec_public_key.hex(),
            "tx_id": self.tx_id.hex(),
            "index": self.index,
            "quantity": self.quantity,
        }

    def __str__(self):
        # TODO: convert EC Public Key to its base58 address
        return "{}(tx_id={}, ec_public_key={}, quantity={})".format(
            self.__class__.__name__, self.tx_id.hex(), self.ec_public_key.hex(), self.quantity,
        )



import struct
from dataclasses import dataclass

import factom_core.primitives as primitives
from factom_core.blockchains import Blockchain
from factom_core.messages import Message


@dataclass
class Heartbeat(Message):
    """
    A heartbeat message for Audit Servers to send out
    """

    TYPE = 10

    timestamp: bytes
    secret_number: int
    height: int
    directory_block_hash: bytes
    chain_id: bytes
    signature: primitives.FullSignature

    def __post_init__(self):
        # TODO: type/value assertions
        self.is_p2p = True
        super().__post_init__()

    def marshal(self) -> bytes:
        """
        Marshal the message into the following representation:
        - first byte is the message type (always 10)
        - next 6 bytes are the timestamp
        - next 4 bytes are a secret number
        - next 4 bytes are the directory block height
        - next 32 bytes are the directory block hash
        - next 32 bytes are the server's identity chain id
        - next 32 bytes are the signing public key
        - next 64 bytes are the signature

        :return: byte representation of the message
        """
        buf = bytearray()
        buf.append(self.TYPE)
        buf.extend(self.timestamp)
        buf.extend(struct.pack(">I", self.secret_number))
        buf.extend(struct.pack(">I", self.height))
        buf.extend(self.chain_id)
        buf.extend(self.signature.marshal())
        return bytes(buf)

    @classmethod
    def unmarshal(cls, raw: bytes):
        msg_type, data = raw[0], raw[1:]
        if msg_type != cls.TYPE:
            raise ValueError("Invalid message type ({})".format(msg_type))

        timestamp, data = data[:6], data[6:]
        secret_number, data = struct.unpack(">I", data[:4])[0], data[4:]
        height, data = struct.unpack(">I", data[:4])[0], data[4:]
        directory_block_hash, data = data[:32], data[32:]
        chain_id, data = data[:32], data[32:]
        signature, data = primitives.FullSignature.unmarshal(data[:96]), data[96:]

        return Heartbeat(
            timestamp=timestamp,
            secret_number=secret_number,
            height=height,
            directory_block_hash=directory_block_hash,
            chain_id=chain_id,
            signature=signature,
        )

    def to_dict(self) -> dict:
        return {
            "timestamp": self.timestamp.hex(),
            "secret_number": self.secret_number,
            "height": self.height,
            "directory_block_hash": self.directory_block_hash.hex(),
            "chain_id": self.chain_id.hex(),
            "signature": self.signature.to_dict(),
        }

    def leader_execute(self, state: Blockchain):
        self.follower_execute(state)

    def follower_execute(self, state: Blockchain):
        pass



import struct
from dataclasses import dataclass


@dataclass
class FullSignature:
    public_key: bytes
    signature: bytes

    def __post_init__(self):
        if type(self.public_key) is not bytes or len(self.public_key) != 32:
            raise ValueError("public_key must be a bytes object of length 32")
        if type(self.signature) is not bytes or len(self.signature) != 64:
            raise ValueError("signature must be a bytes object of length 64")

    def marshal(self):
        buf = bytearray()
        buf.extend(self.public_key)
        buf.extend(self.signature)
        return buf

    @classmethod
    def unmarshal(cls, raw: bytes):
        public_key, data = raw[:32], raw[32:]
        signature, data = data[:64], data[64:]
        return FullSignature(public_key=public_key, signature=signature)

    def to_dict(self):
        return {"public_key": self.public_key.hex(), "signature": self.signature.hex()}


class FullSignatureList(list):
    def marshal(self) -> bytes:
        buf = bytearray()
        buf.extend(struct.pack(">I", len(self)))
        for signature in self:
            buf.extend(signature.marshal())
        return bytes(buf)

    @classmethod
    def unmarshal(cls, raw: bytes):
        length, data = struct.unpack(">I", raw[:4])[0], raw[4:]
        signatures = []
        for i in range(length):
            signature, data = FullSignature.unmarshal(data[:96]), data[96:]
            signatures.append(signature)
        assert len(data) == 0, "Extra bytes remaining!"
        return FullSignatureList(signatures)



import hashlib
import struct
from dataclasses import dataclass, field
from typing import Dict, List

from factom_core.block_elements.factoid_transaction import FactoidTransaction
from factom_core.utils import merkle, varint
from .directory_block import DirectoryBlock


@dataclass
class FactoidBlockHeader:

    CHAIN_ID = bytes.fromhex("000000000000000000000000000000000000000000000000000000000000000f")

    body_mr: bytes
    prev_keymr: bytes
    prev_ledger_keymr: bytes
    ec_exchange_rate: int
    height: int
    expansion_area: bytes
    tx_count: int
    body_size: int

    def __post_init__(self):
        # TODO: value assertions
        pass

    def marshal(self) -> bytes:
        buf = bytearray()
        buf.extend(FactoidBlockHeader.CHAIN_ID)
        buf.extend(self.body_mr)
        buf.extend(self.prev_keymr)
        buf.extend(self.prev_ledger_keymr)
        buf.extend(struct.pack(">Q", self.ec_exchange_rate))
        buf.extend(struct.pack(">I", self.height))
        buf.extend(varint.encode(len(self.expansion_area)))
        buf.extend(self.expansion_area)
        buf.extend(struct.pack(">I", self.tx_count))
        buf.extend(struct.pack(">I", self.body_size))
        return bytes(buf)

    @classmethod
    def unmarshal(cls, raw: bytes):
        h, data = FactoidBlockHeader.unmarshal_with_remainder(raw)
        assert len(data) == 0, "Extra bytes remaining!"
        return h

    @classmethod
    def unmarshal_with_remainder(cls, raw: bytes):
        chain_id, data = raw[:32], raw[32:]
        assert chain_id == FactoidBlockHeader.CHAIN_ID
        body_mr, data = data[:32], data[32:]
        prev_keymr, data = data[:32], data[32:]
        prev_ledger_keymr, data = data[:32], data[32:]
        ec_exchange_rate, data = struct.unpack(">Q", data[:8])[0], data[8:]
        height, data = struct.unpack(">I", data[:4])[0], data[4:]

        header_expansion_size, data = varint.decode(data)
        header_expansion_area, data = (
            data[:header_expansion_size],
            data[header_expansion_size:],
        )

        tx_count, data = struct.unpack(">I", data[:4])[0], data[4:]
        body_size, data = struct.unpack(">I", data[:4])[0], data[4:]
        return (
            FactoidBlockHeader(
                body_mr=body_mr,
                prev_keymr=prev_keymr,
                prev_ledger_keymr=prev_ledger_keymr,
                ec_exchange_rate=ec_exchange_rate,
                height=height,
                expansion_area=header_expansion_area,
                tx_count=tx_count,
                body_size=body_size,
            ),
            data,
        )


@dataclass
class FactoidBlockBody:

    transactions: Dict[int, List[FactoidTransaction]] = field(default_factory=dict)
    _cached_mr: bytes = None

    def __post_init__(self):
        # TODO: value assertions
        pass

    @property
    def merkle_root(self):
        if self._cached_mr is not None:
            return self._cached_mr

        # For Factoid Blocks, body MR is implemented differently in that you first take a single sha256 of every element
        # in the body. And you make a Merkle tree out of the hashed body elements, rather than the elements themselves.
        body_elements = []
        for transactions in self.transactions.values():
            for tx in transactions:
                body_elements.append(tx.hash)
            minute_marker = hashlib.sha256(b"\x00").digest()
            body_elements.append(minute_marker)
        self._cached_mr = merkle.get_merkle_root(body_elements)
        return self._cached_mr

    def marshal(self):
        buf = bytearray()
        for transactions in self.transactions.values():
            for tx in transactions:
                buf.extend(tx.marshal())
            buf.append(0x00)
        return bytes(buf)

    @classmethod
    def unmarshal(cls, raw: bytes, tx_count: int):
        body, data = cls.unmarshal_with_remainder(raw, tx_count)
        assert len(data) == 0, "Extra bytes remaining!"
        return body

    @classmethod
    def unmarshal_with_remainder(cls, raw: bytes, tx_count: int):
        data = raw
        transactions = {}
        current_minute_transactions = []
        minute = 1
        tx_count_observed = 0
        while True:
            if data[0] == 0:
                data = data[1:]
                transactions[minute] = current_minute_transactions
                tx_count_observed += len(current_minute_transactions)
                if minute == 10:
                    break
                current_minute_transactions = []
                minute += 1
                continue
            tx, data = FactoidTransaction.unmarshal_with_remainder(data)
            current_minute_transactions.append(tx)

        assert tx_count_observed == tx_count, "Unexpected transaction count!"

        return FactoidBlockBody(transactions=transactions), data

    def construct_header(
        self, prev_keymr: bytes, prev_ledger_keymr: bytes, ec_exchange_rate: int, height: int,
    ) -> FactoidBlockHeader:
        """
        Seals this factoid block body by constructing and returning it's header
        """
        tx_count = 0
        for tx_list in self.transactions.values():
            tx_count += len(tx_list)
        return FactoidBlockHeader(
            body_mr=self.merkle_root,
            prev_keymr=prev_keymr,
            prev_ledger_keymr=prev_ledger_keymr,
            ec_exchange_rate=ec_exchange_rate,
            height=height,
            expansion_area=b"",
            tx_count=tx_count,
            body_size=len(self.marshal()),
        )


@dataclass
class FactoidBlock:

    header: FactoidBlockHeader
    body: FactoidBlockBody

    _cached_keymr: bytes = None

    def __post_init__(self):
        # TODO: value assertions
        pass

    @property
    def keymr(self):
        if self._cached_keymr is not None:
            return self._cached_keymr

        self._cached_keymr = merkle.calculate_keymr(self.header.marshal(), self.body.merkle_root)
        return self._cached_keymr

    @property
    def ledger_keymr(self):
        pass  # TODO: calculate ledger keymr

    def marshal(self):
        """Marshals the factoid block according to the byte-level representation shown at
        https://github.com/FactomProject/FactomDocs/blob/master/factomDataStructureDetails.md#factoid-block

        Data returned does not include contextual metadata, such as timestamp or the pointer to the next factoid block.
        """
        buf = bytearray()
        buf.extend(self.header.marshal())
        buf.extend(self.body.marshal())
        return bytes(buf)

    @classmethod
    def unmarshal(cls, raw: bytes):
        """Returns a new FactoidBlock object, unmarshalling given bytes according to:
        https://github.com/FactomProject/FactomDocs/blob/master/factomDataStructureDetails.md#factoid-block

        Useful for working with a single fblock out of context, pulled directly from a factomd database for instance.

        FactoidBlock created will not include contextual metadata, such as timestamp or the pointer to the
        next factoid block.
        """
        block, data = cls.unmarshal_with_remainder(raw)
        assert len(data) == 0, "Extra bytes remaining!"
        return block

    @classmethod
    def unmarshal_with_remainder(cls, raw: bytes):
        header, data = FactoidBlockHeader.unmarshal_with_remainder(raw)
        body, data = FactoidBlockBody.unmarshal_with_remainder(data, header.tx_count)
        return FactoidBlock(header=header, body=body), data

    def add_context(self, directory_block: DirectoryBlock):
        pass

    def to_dict(self):
        return {
            "keymr": self.keymr.hex(),
            "body_mr": self.header.body_mr.hex(),
            "prev_keymr": self.header.prev_keymr.hex(),
            "prev_ledger_keymr": self.header.prev_ledger_keymr.hex(),
            "ec_exchange_rate": self.header.ec_exchange_rate,
            "height": self.header.height,
            "expansion_area": self.header.expansion_area.hex(),
            "transaction_count": self.header.tx_count,
            "body_size": self.header.body_size,
            "transactions": {minute: [tx.to_dict() for tx in txs] for minute, txs in self.body.transactions.items()},
        }

    def __str__(self):
        return "{}(height={}, keymr={})".format(self.__class__.__name__, self.header.height, self.keymr.hex())



from .base import Blockchain
from .mainnet import MainnetBlockchain
from .testnet import TestnetBlockchain
from .custom import CustomBlockchain
from .local import LocalBlockchain
from .pending_block import PendingBlock



import datetime
from dataclasses import dataclass, field
from typing import Dict, Union

import factom_core.block_elements as block_elements
import factom_core.blocks as blocks


@dataclass
class PendingBlock:

    admin_block: blocks.AdminBlockBody = field(init=False, default=blocks.AdminBlockBody())
    factoid_block: blocks.FactoidBlockBody = field(init=False, default=blocks.FactoidBlockBody())
    entry_credit_block: blocks.EntryCreditBlockBody = field(init=False, default=blocks.EntryCreditBlockBody())
    entry_blocks: Dict[bytes, blocks.EntryBlockBody] = field(init=False, default_factory=dict)

    previous: blocks.DirectoryBlock = None
    current_minute: int = field(init=False, default=0)
    height: int = field(init=False, default=0)
    timestamp: int = int(datetime.datetime.utcnow().timestamp())

    def __post_init__(self):
        if self.previous is None:
            raise ValueError("PendingBlock must be instantiated with a previous directory block")

    def add_factoid_transaction(self, tx: block_elements.FactoidTransaction):
        if len(self.factoid_block.transactions.keys()) == 0:
            for i in range(1, 10):
                self.factoid_block.transactions[i] = []
        self.factoid_block.transactions[self.current_minute].append(tx)

    def add_commit(self, commit: Union[block_elements.ChainCommit, block_elements.EntryCommit]):
        if self.current_minute not in self.entry_credit_block.objects:
            self.entry_credit_block.objects[self.current_minute] = [commit]
        else:
            self.entry_credit_block.objects[self.current_minute].append(commit)

    def add_entry(self, entry: block_elements.Entry):
        if entry.chain_id not in self.entry_blocks:
            entry_hashes = {self.current_minute: [entry.entry_hash]}
            entry_block = blocks.EntryBlockBody(entry_hashes=entry_hashes)
            self.entry_blocks[entry.chain_id] = entry_block
        elif self.current_minute not in self.entry_blocks[entry.chain_id]:
            self.entry_blocks[entry.chain_id].entry_hashes[self.current_minute] = [entry.entry_hash]
        else:
            self.entry_blocks[entry.chain_id].entry_hashes[self.current_minute].append(entry.entry_hash)



import factom_core.blocks as blocks
from factom_core.blockchains import Blockchain
from factom_core.blockchains.mainnet.constants import MAINNET_NETWORK_ID
from factom_core.blockchains.mainnet.genesis import genesis_factoid_block_bytes


class MainnetBlockchain(Blockchain):
    network_id: bytes = MAINNET_NETWORK_ID

    def load_genesis_block(self) -> blocks.DirectoryBlock:
        body = blocks.AdminBlockBody()
        header = body.construct_header(back_reference_hash=bytes(32), height=0)
        admin_block = blocks.AdminBlock(header, body)

        # Add M1 server index number
        body = blocks.EntryCreditBlockBody(
            objects={1: [0], 2: [], 3: [], 4: [], 5: [], 6: [], 7: [], 8: [], 9: [], 10: [],}
        )
        header = body.construct_header(prev_header_hash=bytes(32), prev_full_hash=bytes(32), height=0)
        entry_credit_block = blocks.EntryCreditBlock(header, body)

        factoid_block = blocks.FactoidBlock.unmarshal(genesis_factoid_block_bytes)

        directory_block_body = blocks.DirectoryBlockBody(
            admin_block_lookup_hash=admin_block.lookup_hash,
            entry_credit_block_header_hash=entry_credit_block.header_hash,
            factoid_block_keymr=factoid_block.keymr,
        )
        directory_block_header = directory_block_body.construct_header(
            network_id=self.network_id, prev_keymr=bytes(32), prev_full_hash=bytes(32), timestamp=24018960, height=0,
        )
        directory_block = blocks.DirectoryBlock(header=directory_block_header, body=directory_block_body)

        # Persist the blocks as new chain heads
        self.db.put_directory_block_head(directory_block)
        self.db.put_admin_block_head(admin_block)
        self.db.put_entry_credit_block_head(entry_credit_block)
        self.db.put_factoid_block_head(factoid_block)

        return directory_block



import struct


def encode(number: int):
    """Pack `number` into varint bytes"""
    buf = bytearray()
    if number == 0:
        buf.append(0x00)

    h = number
    start = False
    if 0x8000000000000000 & h != 0:
        buf.append(0x81)
        start = True

    for i in range(9):
        b = h >> 56
        if b != 0 or start:
            start = True
            b = b | 0x80 if i != 8 else b & 0x7F
            b = b if b < 256 else struct.pack(">Q", b)[-1]
            buf.append(b)

        h = h << 7

    return bytes(buf)


def decode(raw: bytes):
    """Read a varint from `raw` bytes, return the remainder"""
    if raw is None or len(raw) == 0:
        return 0
    result = 0
    data = raw
    while True:
        i, data = ord(data[:1]), data[1:]
        result = result << 7
        result += i & 0x7F
        if i < 0x80:
            break

    return result, data



import struct
from dataclasses import dataclass

import factom_core.primitives as primitives
from factom_core.blockchains import Blockchain
from factom_core.messages import Message
from factom_core.utils import varint


@dataclass
class Ack(Message):
    """
    A acknowledgement for a given message
    """

    TYPE = 1

    vm_index: int
    timestamp: bytes
    salt: bytes
    salt_number: int
    message_hash: bytes
    full_message_hash: bytes
    leader_chain_id: bytes
    height: int
    process_list_height: int
    minute: int
    serial_hash: bytes
    data_area: bytes
    signature: primitives.FullSignature

    def __post_init__(self):
        # TODO: type/value assertions
        self.is_p2p = True
        super().__post_init__()

    def marshal(self) -> bytes:
        """
        Marshal the message into the following representation:
        - first byte is the message type (always 1)
        - next byte is the vm index
        - next 6 bytes are the timestamp
        - next 8 bytes are the salt
        - next 4 bytes are the salt number
        - next 32 bytes are the message hash being acknowledged
        - next 32 bytes are the full message hash
        - next 32 bytes are the leader chain id
        - next 4 bytes are the directory block height
        - next 4 bytes are the process list height TODO: is this actually the process list height?
        - next byte is the minute
        - next 32 bytes are the serial hash
        - next varint bytes are the data area size
        - next X bytes are the data area:
            - TODO: unmarshal the data area in Ack message
        - next 32 bytes are the public key
        - next 64 bytes are the signature

        :return: byte representation of the message
        """
        buf = bytearray()
        buf.append(self.TYPE)
        buf.append(self.vm_index)
        buf.extend(self.timestamp)
        buf.extend(self.salt)
        buf.extend(struct.pack(">I", self.salt_number))
        buf.extend(self.message_hash)
        buf.extend(self.full_message_hash)
        buf.extend(self.leader_chain_id)
        buf.extend(struct.pack(">I", self.height))
        buf.extend(struct.pack(">I", self.process_list_height))
        buf.append(self.minute)
        buf.extend(self.serial_hash)
        buf.extend(varint.encode(len(self.data_area)))
        buf.extend(self.data_area)
        buf.extend(self.signature.marshal())
        return bytes(buf)

    @classmethod
    def unmarshal(cls, raw: bytes):
        msg_type, data = raw[0], raw[1:]
        if msg_type != cls.TYPE:
            raise ValueError("Invalid message type ({})".format(msg_type))

        vm_index, data = data[0], data[1:]
        timestamp, data = data[:6], data[6:]
        salt, data = data[:8], data[8:]
        salt_number, data = struct.unpack(">I", data[:4])[0], data[4:]
        message_hash, data = data[:32], data[32:]
        full_message_hash, data = data[:32], data[32:]
        leader_chain_id, data = data[:32], data[32:]
        height, data = struct.unpack(">I", data[:4])[0], data[4:]
        process_list_height, data = struct.unpack(">I", data[:4])[0], data[4:]
        minute, data = data[0], data[1:]
        serial_hash, data = data[:32], data[32:]
        data_area_size, data = varint.decode(data)
        data_area, data = data[:data_area_size], data[data_area_size:]
        # TODO: unmarshal the data area in Ack message

        signature, data = primitives.FullSignature.unmarshal(data[:96]), data[96:]

        return Ack(
            vm_index=vm_index,
            timestamp=timestamp,
            salt=salt,
            salt_number=salt_number,
            message_hash=message_hash,
            full_message_hash=full_message_hash,
            leader_chain_id=leader_chain_id,
            height=height,
            process_list_height=process_list_height,
            minute=minute,
            serial_hash=serial_hash,
            data_area=data_area,
            signature=signature,
        )

    def to_dict(self):
        return {
            "vm_index": self.vm_index,
            "timestamp": self.timestamp.hex(),
            "salt": self.salt.hex(),
            "salt_number": self.salt_number,
            "message_hash": self.message_hash.hex(),
            "full_message_hash": self.full_message_hash.hex(),
            "leader_chain_id": self.leader_chain_id.hex(),
            "height": self.height,
            "process_list_height": self.process_list_height,
            "minute": self.minute,
            "serial_hash": self.serial_hash.hex(),
            "data_area": self.data_area.hex(),
            "signature": self.signature.to_dict(),
        }

    def leader_execute(self, state: Blockchain):
        self.follower_execute(state)

    def follower_execute(self, state: Blockchain):
        pass



import plyvel
import os
import struct
from typing import List, Tuple, Union

import factom_core.blocks as blocks
import factom_core.block_elements as block_elements

DIRECTORY_BLOCK = b"DirectoryBlock;"
DIRECTORY_BLOCK_NUMBER = b"DirectoryBlockNumber;"
DIRECTORY_BLOCK_SECONDARY = b"DirectoryBlockSecondaryIndex;"

ADMIN_BLOCK = b"AdminBlock;"
ADMIN_BLOCK_NUMBER = b"AdminBlockNumber;"
ADMIN_BLOCK_SECONDARY = b"AdminBlockSecondaryIndex;"

FACTOID_BLOCK = b"FactoidBlock;"
FACTOID_BLOCK_NUMBER = b"FactoidBlockNumber;"
FACTOID_BLOCK_SECONDARY = b"FactoidBlockSecondaryIndex;"

ENTRY_CREDIT_BLOCK = b"EntryCreditBlock;"
ENTRY_CREDIT_BLOCK_NUMBER = b"EntryCreditBlockNumber;"
ENTRY_CREDIT_BLOCK_SECONDARY = b"EntryCreditBlockSecondaryIndex;"

CHAIN_HEAD = b"ChainHead;"

ENTRY_BLOCK = b"EntryBlock;"
ENTRY_BLOCK_NUMBER = b"EntryBlockNumber;"
ENTRY_BLOCK_SECONDARY = b"EntryBlockSecondaryIndex;"

ENTRY = b"Entry;"

DIR_BLOCK_INFO = b"DirBlockInfo;"
DIR_BLOCK_INFO_UNCONFIRMED = b"DirBlockInfoUnconfirmed;"
DIR_BLOCK_INFO_NUMBER = b"DirBlockInfoNumber;"
DIR_BLOCK_INFO_SECONDARY = b"DirBlockInfoSecondaryIndex;"

INCLUDED_IN = b"IncludedIn;"

PAID_FOR = b"PaidFor;"
KEY_VALUE_STORE = b"KeyValueStore;"


FullBlockSet = Tuple[
    blocks.DirectoryBlock, blocks.AdminBlock, blocks.EntryCreditBlock, blocks.FactoidBlock, List[blocks.EntryBlock],
]


class FactomdLevelDB:
    def __init__(self, path: str = None, **kwargs):
        """
        A wrapper around the legacy factomd level-db

        :param path: filepath to the factomd leveldb database, defaults to: /$HOME/.factom/hydra/data/
        """
        if path is None:
            home = os.getenv("HOME")
            path = f"{home}/.factom/hydra/data/"
        self._db = plyvel.DB(path, **kwargs)

    def close(self):
        self._db.close()

    def get_chain_head(self, chain_id: bytes):
        sub_db = self._db.prefixed_db(CHAIN_HEAD)
        return sub_db.get(chain_id)

    def put_chain_head(self, chain_id: bytes, head: bytes):
        sub_db = self._db.prefixed_db(CHAIN_HEAD)
        sub_db.put(chain_id, head)

    #
    # Directory Block
    #

    def get_directory_block(self, **kwargs) -> Union[blocks.DirectoryBlock, None]:
        keymr = kwargs.get("keymr")
        height = kwargs.get("height")
        assert (keymr is None and type(height) is int) or (type(keymr) is bytes and height is None)
        if height is not None:
            sub_db = self._db.prefixed_db(DIRECTORY_BLOCK_NUMBER)
            height_encoded = struct.pack(">I", height)
            keymr = sub_db.get(height_encoded)
            if keymr is None:
                return None
        sub_db = self._db.prefixed_db(DIRECTORY_BLOCK)
        raw = sub_db.get(keymr)
        return None if raw is None else blocks.DirectoryBlock.unmarshal(raw)

    def get_directory_block_head(self) -> Union[blocks.DirectoryBlock, None]:
        prev_keymr = self.get_chain_head(blocks.DirectoryBlockHeader.CHAIN_ID)
        if prev_keymr is None:
            return None
        return self.get_directory_block(keymr=prev_keymr)

    def put_directory_block(self, block: blocks.DirectoryBlock):
        sub_db = self._db.prefixed_db(DIRECTORY_BLOCK_NUMBER)
        sub_db.put(struct.pack(">I", block.header.height), block.keymr)
        sub_db = self._db.prefixed_db(DIRECTORY_BLOCK)
        sub_db.put(block.keymr, block.marshal())

    def put_directory_block_head(self, block: blocks.DirectoryBlock):
        self.put_directory_block(block)
        self.put_chain_head(block.header.CHAIN_ID, block.keymr)

    #
    # Admin Block
    #

    def get_admin_block(self, **kwargs) -> Union[blocks.AdminBlock, None]:
        lookup_hash = kwargs.get("lookup_hash")
        height = kwargs.get("height")
        assert (lookup_hash is None and type(height) is int) or (type(lookup_hash) is bytes and height is None)
        if height is not None:
            sub_db = self._db.prefixed_db(ADMIN_BLOCK_NUMBER)
            height_encoded = struct.pack(">I", height)
            lookup_hash = sub_db.get(height_encoded)
            if lookup_hash is None:
                return None
        sub_db = self._db.prefixed_db(ADMIN_BLOCK)
        raw = sub_db.get(lookup_hash)
        if raw is None:
            return None
        block = blocks.AdminBlock.unmarshal(raw)
        return block

    def get_admin_block_head(self) -> Union[blocks.AdminBlock, None]:
        prev_hash = self.get_chain_head(blocks.AdminBlockHeader.CHAIN_ID)
        if prev_hash is None:
            return None
        return self.get_admin_block(lookup_hash=prev_hash)

    def put_admin_block(self, block: blocks.AdminBlock):
        sub_db = self._db.prefixed_db(ADMIN_BLOCK_NUMBER)
        sub_db.put(struct.pack(">I", block.header.height), block.lookup_hash)
        sub_db = self._db.prefixed_db(ADMIN_BLOCK)
        sub_db.put(block.lookup_hash, block.marshal())

    def put_admin_block_head(self, block: blocks.AdminBlock):
        self.put_admin_block(block)
        self.put_chain_head(block.header.CHAIN_ID, block.lookup_hash)

    #
    # Factoid Block
    #

    def get_factoid_block(self, **kwargs) -> Union[blocks.FactoidBlock, None]:
        keymr = kwargs.get("keymr")
        height = kwargs.get("height")
        assert (keymr is None and type(height) is int) or (type(keymr) is bytes and height is None)
        if height is not None:
            sub_db = self._db.prefixed_db(FACTOID_BLOCK_NUMBER)
            height_encoded = struct.pack(">I", height)
            keymr = sub_db.get(height_encoded)
            if keymr is None:
                return None
        sub_db = self._db.prefixed_db(FACTOID_BLOCK)
        raw = sub_db.get(keymr)
        if raw is None:
            return None
        block = blocks.FactoidBlock.unmarshal(raw)
        return block

    def get_factoid_block_head(self) -> Union[blocks.FactoidBlock, None]:
        prev_keymr = self.get_chain_head(blocks.FactoidBlockHeader.CHAIN_ID)
        if prev_keymr is None:
            return None
        return self.get_factoid_block(keymr=prev_keymr)

    def put_factoid_block(self, block: blocks.FactoidBlock):
        sub_db = self._db.prefixed_db(FACTOID_BLOCK_NUMBER)
        sub_db.put(struct.pack(">I", block.header.height), block.keymr)
        sub_db = self._db.prefixed_db(FACTOID_BLOCK)
        sub_db.put(block.keymr, block.marshal())

    def put_factoid_block_head(self, block: blocks.FactoidBlock):
        self.put_factoid_block(block)
        self.put_chain_head(block.header.CHAIN_ID, block.keymr)

    #
    # Entry Credit Block
    #

    def get_entry_credit_block(self, **kwargs) -> Union[blocks.EntryCreditBlock, None]:
        header_hash = kwargs.get("header_hash")
        height = kwargs.get("height")
        assert (header_hash is None and type(height) is int) or (type(header_hash) is bytes and height is None)
        if height is not None:
            sub_db = self._db.prefixed_db(ENTRY_CREDIT_BLOCK_NUMBER)
            height_encoded = struct.pack(">I", height)
            header_hash = sub_db.get(height_encoded)
            if header_hash is None:
                return None
        sub_db = self._db.prefixed_db(ENTRY_CREDIT_BLOCK)
        raw = sub_db.get(header_hash)
        if raw is None:
            return None
        block = blocks.EntryCreditBlock.unmarshal(raw)
        return block

    def get_entry_credit_block_head(self) -> Union[blocks.EntryCreditBlock, None]:
        prev_hash = self.get_chain_head(blocks.EntryCreditBlockHeader.CHAIN_ID)
        if prev_hash is None:
            return None
        return self.get_entry_credit_block(header_hash=prev_hash)

    def put_entry_credit_block(self, block: blocks.EntryCreditBlock):
        sub_db = self._db.prefixed_db(ENTRY_CREDIT_BLOCK_NUMBER)
        sub_db.put(struct.pack(">I", block.header.height), block.header_hash)
        sub_db = self._db.prefixed_db(ENTRY_CREDIT_BLOCK)
        sub_db.put(block.header_hash, block.marshal())

    def put_entry_credit_block_head(self, block: blocks.EntryCreditBlock):
        self.put_entry_credit_block(block)
        self.put_chain_head(block.header.CHAIN_ID, block.header_hash)

    #
    # Entry Block
    #

    def get_entry_block(self, keymr: bytes) -> Union[blocks.EntryBlock, None]:
        sub_db = self._db.prefixed_db(ENTRY_BLOCK)
        raw = sub_db.get(keymr)
        if raw is None:
            return None
        block = blocks.EntryBlock.unmarshal(raw)
        return block

    def get_entry_block_head(self, chain_id: bytes) -> Union[blocks.EntryBlock, None]:
        prev_keymr = self.get_chain_head(chain_id)
        if prev_keymr is None:
            return None
        return self.get_entry_block(prev_keymr)

    def put_entry_block(self, block: blocks.EntryBlock):
        sub_db = self._db.prefixed_db(ENTRY_BLOCK)
        sub_db.put(block.keymr, block.marshal())

    def put_entry_block_head(self, block: blocks.EntryBlock):
        self.put_entry_block(block)
        self.put_chain_head(block.header.chain_id, block.keymr)

    #
    # Entry
    #

    def get_entry(self, entry_hash: bytes) -> block_elements.Entry:
        sub_db = self._db.prefixed_db(ENTRY)
        chain_id = sub_db.get(entry_hash)
        if chain_id is None:
            return None
        raw = self._db.get(chain_id + b";" + entry_hash)
        return None if raw is None else block_elements.Entry.unmarshal(raw)

    def put_entry(self, entry: block_elements.Entry):
        sub_db = self._db.prefixed_db(ENTRY)
        sub_db.put(entry.entry_hash, entry.chain_id)
        sub_db = self._db.prefixed_db(entry.chain_id + ";".encode())
        sub_db.put(entry.entry_hash, entry.marshal())



import hashlib
from dataclasses import dataclass, field
from typing import List

from factom_core.block_elements.admin_messages import *
from factom_core.utils import varint
from .directory_block import DirectoryBlock


@dataclass
class AdminBlockHeader:

    CHAIN_ID = bytes.fromhex("000000000000000000000000000000000000000000000000000000000000000a")

    prev_back_reference_hash: bytes
    height: int
    expansion_area: bytes
    message_count: int
    body_size: int

    def __post_init__(self):
        # TODO: value assertions
        pass

    def marshal(self) -> bytes:
        buf = bytearray()
        buf.extend(AdminBlockHeader.CHAIN_ID)
        buf.extend(self.prev_back_reference_hash)
        buf.extend(struct.pack(">I", self.height))
        buf.extend(varint.encode(len(self.expansion_area)))
        buf.extend(self.expansion_area)
        buf.extend(struct.pack(">I", self.message_count))
        buf.extend(struct.pack(">I", self.body_size))
        return bytes(buf)

    @classmethod
    def unmarshal(cls, raw: bytes):
        h, data = AdminBlockHeader.unmarshal_with_remainder(raw)
        assert len(data) == 0, "Extra bytes remaining!"
        return h

    @classmethod
    def unmarshal_with_remainder(cls, raw: bytes):
        chain_id, data = raw[:32], raw[32:]
        assert chain_id == AdminBlockHeader.CHAIN_ID
        prev_back_reference_hash, data = data[:32], data[32:]
        height, data = struct.unpack(">I", data[:4])[0], data[4:]

        expansion_size, data = varint.decode(data)
        expansion_area, data = data[:expansion_size], data[expansion_size:]
        # TODO: unmarshal header expansion area

        message_count, data = struct.unpack(">I", data[:4])[0], data[4:]
        body_size, data = struct.unpack(">I", data[:4])[0], data[4:]
        return (
            AdminBlockHeader(
                prev_back_reference_hash=prev_back_reference_hash,
                height=height,
                expansion_area=expansion_area,
                message_count=message_count,
                body_size=body_size,
            ),
            data,
        )


@dataclass
class AdminBlockBody:

    messages: List[AdminMessage] = field(default_factory=list)

    def __post_init__(self):
        # TODO: value assertions
        pass

    def marshal(self) -> bytes:
        buf = bytearray()
        for message in self.messages:
            if type(message) is int:
                buf.append(message)
                continue
            buf.append(message.__class__.ADMIN_ID)
            buf.extend(message.marshal())
        return bytes(buf)

    @classmethod
    def unmarshal(cls, raw: bytes, message_count: int):
        body, data = cls.unmarshal_with_remainder(raw, message_count)
        assert len(data) == 0, "Extra bytes remaining!"
        return body

    @classmethod
    def unmarshal_with_remainder(cls, raw: bytes, message_count: int):
        data = raw
        messages = []
        for i in range(message_count):
            admin_id, data = data[0], data[1:]
            msg = None
            if admin_id == MinuteNumber.ADMIN_ID:  # Deprecated in M2
                size = MinuteNumber.MESSAGE_SIZE
                msg_data, data = data[:size], data[size:]
                msg = MinuteNumber.unmarshal(msg_data)

            elif admin_id == DirectoryBlockSignature.ADMIN_ID:
                size = DirectoryBlockSignature.MESSAGE_SIZE
                msg_data, data = data[:size], data[size:]
                msg = DirectoryBlockSignature.unmarshal(msg_data)

            elif admin_id == MatryoshkaHashReveal.ADMIN_ID:
                size = MatryoshkaHashReveal.MESSAGE_SIZE
                msg_data, data = data[:size], data[size:]
                msg = MatryoshkaHashReveal.unmarshal(msg_data)

            elif admin_id == MatryoshkaHashAddOrReplace.ADMIN_ID:
                size = MatryoshkaHashAddOrReplace.MESSAGE_SIZE
                msg_data, data = data[:size], data[size:]
                msg = MatryoshkaHashAddOrReplace.unmarshal(msg_data)

            elif admin_id == ServerCountIncrease.ADMIN_ID:
                size = ServerCountIncrease.MESSAGE_SIZE
                msg_data, data = data[:size], data[size:]
                msg = ServerCountIncrease.unmarshal(msg_data)

            elif admin_id == AddFederatedServer.ADMIN_ID:
                size = AddFederatedServer.MESSAGE_SIZE
                msg_data, data = data[:size], data[size:]
                msg = AddFederatedServer.unmarshal(msg_data)

            elif admin_id == AddAuditServer.ADMIN_ID:
                size = AddAuditServer.MESSAGE_SIZE
                msg_data, data = data[:size], data[size:]
                msg = AddAuditServer.unmarshal(msg_data)

            elif admin_id == RemoveFederatedServer.ADMIN_ID:
                size = RemoveFederatedServer.MESSAGE_SIZE
                msg_data, data = data[:size], data[size:]
                msg = RemoveFederatedServer.unmarshal(msg_data)

            elif admin_id == AddFederatedServerSigningKey.ADMIN_ID:
                size = AddFederatedServerSigningKey.MESSAGE_SIZE
                msg_data, data = data[:size], data[size:]
                msg = AddFederatedServerSigningKey.unmarshal(msg_data)

            elif admin_id == AddFederatedServerBitcoinAnchorKey.ADMIN_ID:
                size = AddFederatedServerBitcoinAnchorKey.MESSAGE_SIZE
                msg_data, data = data[:size], data[size:]
                msg = AddFederatedServerBitcoinAnchorKey.unmarshal(msg_data)

            elif admin_id == ServerFaultHandoff.ADMIN_ID:
                msg = ServerFaultHandoff()  # No data on chain for message

            elif admin_id == CoinbaseDescriptor.ADMIN_ID:
                msg, data = CoinbaseDescriptor.unmarshal_with_remainder(data)

            elif admin_id == CoinbaseDescriptorCancel.ADMIN_ID:
                msg, data = CoinbaseDescriptorCancel.unmarshal_with_remainder(data)

            elif admin_id == AddAuthorityFactoidAddress.ADMIN_ID:
                size = AddAuthorityFactoidAddress.MESSAGE_SIZE
                msg_data, data = data[:size], data[size:]
                msg = AddAuthorityFactoidAddress.unmarshal(msg_data)

            elif admin_id == AddAuthorityEfficiency.ADMIN_ID:
                size = AddAuthorityEfficiency.MESSAGE_SIZE
                msg_data, data = data[:size], data[size:]
                msg = AddAuthorityFactoidAddress.unmarshal(msg_data)

            elif admin_id <= 0x0E:
                msg = admin_id
                print(f"Unsupported admin message type {admin_id} found")

            if msg is not None:
                messages.append(msg)

        assert len(messages) == message_count, "Unexpected message count"

        return AdminBlockBody(messages=messages), data

    def construct_header(self, prev_back_reference_hash: bytes, height: int) -> AdminBlockHeader:
        """
        Seals the admin block by constructing and returning its header
        """
        return AdminBlockHeader(
            prev_back_reference_hash=prev_back_reference_hash,
            height=height,
            expansion_area=b"",
            message_count=len(self.messages),
            body_size=len(self.marshal()),
        )


@dataclass
class AdminBlock:

    header: AdminBlockHeader
    body: AdminBlockBody

    _cached_lookup_hash: bytes = None
    _cached_back_reference_hash: bytes = None

    def __post_init__(self):
        # TODO: value assertions
        pass

    @property
    def lookup_hash(self):
        if self._cached_lookup_hash is not None:
            return self._cached_lookup_hash
        self._cached_lookup_hash = hashlib.sha256(self.marshal()).digest()
        return self._cached_lookup_hash

    @property
    def back_reference_hash(self):
        if self._cached_back_reference_hash is not None:
            return self._cached_back_reference_hash
        self._cached_back_reference_hash = hashlib.sha512(self.marshal()).digest()[:32]
        return self._cached_back_reference_hash

    def marshal(self) -> bytes:
        buf = bytearray()
        buf.extend(self.header.marshal())
        buf.extend(self.body.marshal())
        return bytes(buf)

    @classmethod
    def unmarshal(cls, raw: bytes):
        """Returns a new AdminBlock object, unmarshalling given bytes according to:
        https://github.com/FactomProject/FactomDocs/blob/master/factomDataStructureDetails.md#administrative-block

        Useful for working with a single ablock out of context, pulled directly from a factomd database for instance.

        AdminBlock created will not include contextual metadata, such as timestamp
        """
        block, data = cls.unmarshal_with_remainder(raw)
        assert len(data) == 0, "Extra bytes remaining!"
        return block

    @classmethod
    def unmarshal_with_remainder(cls, raw: bytes):
        header, data = AdminBlockHeader.unmarshal_with_remainder(raw)

        messages = []
        for i in range(header.message_count):
            admin_id, data = data[0], data[1:]
            msg = None
            if admin_id == MinuteNumber.ADMIN_ID:  # Deprecated in M2
                size = MinuteNumber.MESSAGE_SIZE
                msg_data, data = data[:size], data[size:]
                msg = MinuteNumber.unmarshal(msg_data)

            elif admin_id == DirectoryBlockSignature.ADMIN_ID:
                size = DirectoryBlockSignature.MESSAGE_SIZE
                msg_data, data = data[:size], data[size:]
                msg = DirectoryBlockSignature.unmarshal(msg_data)

            elif admin_id == MatryoshkaHashReveal.ADMIN_ID:
                size = MatryoshkaHashReveal.MESSAGE_SIZE
                msg_data, data = data[:size], data[size:]
                msg = MatryoshkaHashReveal.unmarshal(msg_data)

            elif admin_id == MatryoshkaHashAddOrReplace.ADMIN_ID:
                size = MatryoshkaHashAddOrReplace.MESSAGE_SIZE
                msg_data, data = data[:size], data[size:]
                msg = MatryoshkaHashAddOrReplace.unmarshal(msg_data)

            elif admin_id == ServerCountIncrease.ADMIN_ID:
                size = ServerCountIncrease.MESSAGE_SIZE
                msg_data, data = data[:size], data[size:]
                msg = ServerCountIncrease.unmarshal(msg_data)

            elif admin_id == AddFederatedServer.ADMIN_ID:
                size = AddFederatedServer.MESSAGE_SIZE
                msg_data, data = data[:size], data[size:]
                msg = AddFederatedServer.unmarshal(msg_data)

            elif admin_id == AddAuditServer.ADMIN_ID:
                size = AddAuditServer.MESSAGE_SIZE
                msg_data, data = data[:size], data[size:]
                msg = AddAuditServer.unmarshal(msg_data)

            elif admin_id == RemoveFederatedServer.ADMIN_ID:
                size = RemoveFederatedServer.MESSAGE_SIZE
                msg_data, data = data[:size], data[size:]
                msg = RemoveFederatedServer.unmarshal(msg_data)

            elif admin_id == AddFederatedServerSigningKey.ADMIN_ID:
                size = AddFederatedServerSigningKey.MESSAGE_SIZE
                msg_data, data = data[:size], data[size:]
                msg = AddFederatedServerSigningKey.unmarshal(msg_data)

            elif admin_id == AddFederatedServerBitcoinAnchorKey.ADMIN_ID:
                size = AddFederatedServerBitcoinAnchorKey.MESSAGE_SIZE
                msg_data, data = data[:size], data[size:]
                msg = AddFederatedServerBitcoinAnchorKey.unmarshal(msg_data)

            elif admin_id == ServerFaultHandoff.ADMIN_ID:
                msg = ServerFaultHandoff()  # No data on chain for message

            elif admin_id == CoinbaseDescriptor.ADMIN_ID:
                msg, data = CoinbaseDescriptor.unmarshal_with_remainder(data)

            elif admin_id == CoinbaseDescriptorCancel.ADMIN_ID:
                msg, data = CoinbaseDescriptorCancel.unmarshal_with_remainder(data)

            elif admin_id == AddAuthorityFactoidAddress.ADMIN_ID:
                size = AddAuthorityFactoidAddress.MESSAGE_SIZE
                msg_data, data = data[:size], data[size:]
                msg = AddAuthorityFactoidAddress.unmarshal(msg_data)

            elif admin_id == AddAuthorityEfficiency.ADMIN_ID:
                size = AddAuthorityEfficiency.MESSAGE_SIZE
                msg_data, data = data[:size], data[size:]
                msg = AddAuthorityFactoidAddress.unmarshal(msg_data)

            elif admin_id <= 0x0E:
                msg = admin_id
                print("Unsupported admin message type {} found at Admin Block {}".format(admin_id, header.height))

            if msg is not None:
                messages.append(msg)

        assert len(messages) == header.message_count, "Unexpected message count at Admin Block {}".format(header.height)

        body = AdminBlockBody(messages=messages)
        return AdminBlock(header=header, body=body), data

    def add_context(self, directory_block: DirectoryBlock):
        pass

    def to_dict(self):
        return {
            "lookup_hash": self.lookup_hash.hex(),
            "back_reference_hash": self.header.prev_back_reference_hash.hex(),
            "height": self.header.height,
            "expansion_area": self.header.expansion_area.hex(),
            "message_count": self.header.message_count,
            "body_size": self.header.body_size,
            "messages": [m.to_dict() for m in self.body.messages],
        }

    def __str__(self):
        return "{}(height={}, hash={})".format(self.__class__.__name__, self.header.height, self.lookup_hash.hex())



import hashlib
import struct
from dataclasses import dataclass, field
from typing import Dict, List, Union

from factom_core.block_elements.balance_increase import BalanceIncrease
from factom_core.block_elements.chain_commit import ChainCommit
from factom_core.block_elements.entry_commit import EntryCommit
from factom_core.utils import varint
from .directory_block import DirectoryBlock


ECIDTypes = Union[ChainCommit, EntryCommit, int]


@dataclass
class EntryCreditBlockHeader:

    CHAIN_ID = bytes.fromhex("000000000000000000000000000000000000000000000000000000000000000c")

    body_hash: bytes
    prev_header_hash: bytes
    prev_full_hash: bytes
    height: int
    expansion_area: bytes
    object_count: int
    body_size: int

    def __post_init__(self):
        # TODO: value assertions
        pass

    def marshal(self) -> bytes:
        buf = bytearray()
        buf.extend(EntryCreditBlockHeader.CHAIN_ID)
        buf.extend(self.body_hash)
        buf.extend(self.prev_header_hash)
        buf.extend(self.prev_full_hash)
        buf.extend(struct.pack(">I", self.height))
        buf.extend(varint.encode(len(self.expansion_area)))
        buf.extend(self.expansion_area)
        buf.extend(struct.pack(">Q", self.object_count))
        buf.extend(struct.pack(">Q", self.body_size))
        return bytes(buf)

    @classmethod
    def unmarshal(cls, raw: bytes):
        h, data = EntryCreditBlockHeader.unmarshal_with_remainder(raw)
        assert len(data) == 0, "Extra bytes remaining!"
        return h

    @classmethod
    def unmarshal_with_remainder(cls, raw: bytes):
        chain_id, data = raw[:32], raw[32:]
        assert chain_id == EntryCreditBlockHeader.CHAIN_ID
        body_hash, data = data[:32], data[32:]
        prev_header_hash, data = data[:32], data[32:]
        prev_full_hash, data = data[:32], data[32:]
        height, data = struct.unpack(">I", data[:4])[0], data[4:]

        header_expansion_size, data = varint.decode(data)
        header_expansion_area, data = (
            data[:header_expansion_size],
            data[header_expansion_size:],
        )

        object_count, data = struct.unpack(">Q", data[:8])[0], data[8:]
        body_size, data = struct.unpack(">Q", data[:8])[0], data[8:]

        return (
            EntryCreditBlockHeader(
                body_hash=body_hash,
                prev_header_hash=prev_header_hash,
                prev_full_hash=prev_full_hash,
                height=height,
                expansion_area=header_expansion_area,
                object_count=object_count,
                body_size=body_size,
            ),
            data,
        )


@dataclass
class EntryCreditBlockBody:

    objects: Dict[int, List[ECIDTypes]] = field(default_factory=dict)

    def __post_init__(self):
        # TODO: value assertions
        pass

    def marshal(self):
        buf = bytearray()
        for minute, objects in self.objects.items():
            for o in objects:
                if isinstance(o, int):
                    buf.append(0x00)
                    buf.append(o)
                elif isinstance(o, ChainCommit):
                    buf.append(ChainCommit.ECID)
                    buf.extend(o.marshal())
                elif isinstance(o, EntryCommit):
                    buf.append(EntryCommit.ECID)
                    buf.extend(o.marshal())
                elif isinstance(o, BalanceIncrease):
                    buf.append(BalanceIncrease.ECID)
                    buf.extend(o.marshal())
                else:
                    raise ValueError("Invalid ECID type!")
            buf.append(0x01)
            buf.append(minute)
        return bytes(buf)

    @classmethod
    def unmarshal(cls, raw: bytes, object_count: int):
        body, data = cls.unmarshal_with_remainder(raw, object_count)
        assert len(data) == 0, "Extra bytes remaining!"
        return body

    @classmethod
    def unmarshal_with_remainder(cls, raw: bytes, object_count: int):
        data = raw
        objects = {}  # map of minute --> objects array
        current_minute_objects = []
        for i in range(object_count):
            ecid, data = data[0], data[1:]
            if ecid == 0x00:
                server_index, data = data[0], data[1:]
                current_minute_objects.append(server_index)
            elif ecid == 0x01:
                minute, data = data[0], data[1:]
                objects[minute] = current_minute_objects
                current_minute_objects = []
            elif ecid == ChainCommit.ECID:
                chain_commit, data = (
                    data[: ChainCommit.BITLENGTH],
                    data[ChainCommit.BITLENGTH :],
                )
                chain_commit = ChainCommit.unmarshal(chain_commit)
                current_minute_objects.append(chain_commit)
            elif ecid == EntryCommit.ECID:
                entry_commit, data = (
                    data[: EntryCommit.BITLENGTH],
                    data[EntryCommit.BITLENGTH :],
                )
                entry_commit = EntryCommit.unmarshal(entry_commit)
                current_minute_objects.append(entry_commit)
            elif ecid == BalanceIncrease.ECID:
                balance_increase, data = BalanceIncrease.unmarshal_with_remainder(data)
                current_minute_objects.append(balance_increase)
            else:
                raise ValueError

        return EntryCreditBlockBody(objects=objects), data

    def construct_header(self, prev_header_hash: bytes, prev_full_hash: bytes, height: int) -> EntryCreditBlockHeader:
        object_count = 0
        for object_list in self.objects.values():
            object_count += len(object_list) + 1
        marshalled_body = self.marshal()
        return EntryCreditBlockHeader(
            body_hash=hashlib.sha256(marshalled_body).digest(),
            prev_header_hash=prev_header_hash,
            prev_full_hash=prev_full_hash,
            height=height,
            expansion_area=b"",
            object_count=object_count,
            body_size=len(marshalled_body),
        )


@dataclass
class EntryCreditBlock:

    header: EntryCreditBlockHeader
    body: EntryCreditBlockBody

    _cached_header_hash: bytes = None

    def __post_init__(self):
        # TODO: value assertions
        pass

    @property
    def header_hash(self):
        if self._cached_header_hash is not None:
            return self._cached_header_hash
        self._cached_header_hash = hashlib.sha256(self.header.marshal()).digest()
        return self._cached_header_hash

    @property
    def full_hash(self):
        return hashlib.sha256(self.marshal()).digest()

    def marshal(self):
        """Marshals the directory block according to the byte-level representation shown at
        https://github.com/FactomProject/FactomDocs/blob/master/factomDataStructureDetails.md#entry-credit-block

        Data returned does not include contextual metadata, such as timestamp or the pointer to the
        next entry-credit block.
        """
        buf = bytearray()
        buf.extend(self.header.marshal())
        buf.extend(self.body.marshal())
        return bytes(buf)

    @classmethod
    def unmarshal(cls, raw: bytes):
        """Returns a new EntryCreditBlock object, unmarshalling given bytes according to:
        https://github.com/FactomProject/FactomDocs/blob/master/factomDataStructureDetails.md#entry-credit-block

        Useful for working with a single ecblock out of context, pulled directly from a factomd database for instance.

        EntryCreditBlock created will not include contextual metadata, such as timestamp or the pointer to the
        next entry-credit block.
        """
        block, data = cls.unmarshal_with_remainder(raw)
        assert len(data) == 0, "Extra bytes remaining!"
        return block

    @classmethod
    def unmarshal_with_remainder(cls, raw: bytes):
        header, data = EntryCreditBlockHeader.unmarshal_with_remainder(raw)
        body, data = EntryCreditBlockBody.unmarshal_with_remainder(data, header.object_count)
        return EntryCreditBlock(header=header, body=body), data

    def add_context(self, directory_block: DirectoryBlock):
        pass

    def to_dict(self):
        return {
            "header_hash": self.header_hash.hex(),
            "body_hash": self.header.body_hash.hex(),
            "prev_header_hash": self.header.prev_header_hash.hex(),
            "prev_full_hash": self.header.prev_full_hash.hex(),
            "height": self.header.height,
            "expansion_area": self.header.expansion_area.hex(),
            "object_count": self.header.object_count,
            "body_size": self.header.body_size,
            "objects": {
                minute: [o if type(o) is int else o.to_dict() for o in objects]
                for minute, objects in self.body.objects.items()
            },
        }

    def __str__(self):
        return "{}(height={})".format(self.__class__.__name__, self.header.height)
